/**
 * Filelist module for the Pylon IDE
 *
 * @license GPLv3 <http://www.gnu.org/licenses/gpl.txt>
 */

"use strict";

var Url = require("url");
var Plugin = require("../pylon.core/plugin");
var FilelistLib = require("./filelist");
var util = require("util");
var error = require("http-error");

var name = "filelist";

module.exports = function setup(options, imports, register) {
    var Filelist = new FilelistLib();
    Filelist.setEnv({
        findCmd: options.findCmd,
        platform: options.platform
    });

    var Vfs = imports["vfs"];
    var IdeRoutes = imports["ide-routes"];
    var Permissions = imports["workspace-permissions"];

    var FilelistPlugin = function(ide, workspace) {
        Plugin.call(this, ide, workspace);
        this.name = name;
        this.processCount = 0;
        this.ws = ide.workspace.workspaceId;

        Filelist.setEnv({
            workspaceId: this.ws,
            basePath: ide.workspace.workspaceDir
        });

        // set up some routes as well
        var self = this;
        IdeRoutes.use("/fs", imports.connect.getRouter()(function(app) {
            app.get("/list", self.getList.bind(self));
        }));
    };

    util.inherits(FilelistPlugin, Plugin);

    (function() {

        this.getList = function(req, res, next) {
            var uid = req.session.uid || req.session.anonid;
            if (!uid)
                return next(new error.Unauthorized());

            // does this user has read-permissions...
            Permissions.getPermissions(uid, this.ws, "fs_filelist", function(err, perms) {
                if (err)
                    return next(err);

                if (perms.fs.indexOf("r") === -1)
                    return next(new error.Forbidden("You are not allowed to view this resource"));

                // and kick off the download action!
                var query = Url.parse(req.url, true).query;
                query.showHiddenFiles = !!parseInt(query.showHiddenFiles, 10);

                Filelist.exec(query, Vfs,
                    // incoming data
                    function(msg) {
                        if (!msg)
                            return;

                        if (!res.headersSent)
                            res.writeHead(200, { "content-type": "text/plain" });
                        res.write(msg);
                    },
                    // process exit
                    function(code, stderr) {
                        if (code === 0)
                            return res.end();

                        var errMsg = "Process terminated with code " + code + ", " + stderr;
                        if (!res.headerSent)
                            return next(errMsg);

                        console.error(errMsg);
                        res.end();
                    }
                );
            });
        };

    }).call(FilelistPlugin.prototype);

    imports.ide.register(name, FilelistPlugin, register);
};
