/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 *
 */

// #ifdef __AMLSPINNER || __INC_ALL
/** 
 * This element is used to choose a number via plus/minus buttons.
 *
 * When the plus button is clicked/held longer, the number increments faster. The same
 * situation occurs for the minus button. It's also possible to increment and decrement
 * value by moving mouse cursor up or down with clicked input. 
 * 
 * Max and min attributes define the range of allowed values.
 * 
 * #### Example: Setting Maximum and Minimum Ranges
 *
 * ```xml, demo
 * <a:application xmlns:a="https://github.com/pylonide/pylon">
 *  <!-- startcontent -->
 *    <a:spinner value="6" min="-6" max="12" width="200"></a:spinner>
 *  <!-- endcontent -->
 * </a:application>
 * ```
 * 
 * #### Example: Loading Data
 * 
 * ```xml, demo
 * <a:application xmlns:a="https://github.com/pylonide/pylon">
 *   <!-- startcontent -->
 *   <a:model id="mdlSpinner">
 *       <data value="56"></data>
 *   </a:model>
 *   <a:spinner value="[@value]" model="mdlSpinner" />
 *   <!-- endcontent -->
 * </a:application>
 * ```
 * 
 * #### Example: Connecting to a Textbox
 *
 * 
 * ```xml, demo
 * <a:application xmlns:a="https://github.com/pylonide/pylon">
 *   <!-- startcontent -->
 *   <a:model id="mdlTest">
 *     <overview page="1" pages="10" />
 *   </a:model>
 *   <a:spinner 
 *     id      = "spinner" 
 *     min     = "0" 
 *     max     = "[@pages]" 
 *     model   = "mdlTest" 
 *     value   = "[@page]" 
 *     caption = "[@page] of [@pages]">
 *   </a:spinner>
 *   <a:textbox value="{spinner.caption}"></a:textbox>
 *   <!-- endcontent -->
 * </a:application>
 * ```
 *
 * @class ppc.spinner
 * @define spinner
 * @form
 * @version     %I%, %G%
 * 
 * @inherits ppc.StandardBinding
 * @inherits ppc.DataAction
 * @inheritsElsewhere ppc.XForms
 *
 */
/**
 * @attribute {Number}   [max=64000]       Sets or gets the maximum allowed value
 */
/**
 * @attribute {Number}   [min=-64000]      Sets or gets the minimal allowed value
 */
/**
 *  @attribute {Number}   value     Sets or gets the actual value displayed in component
 * 
 */
/**
 * @binding value  Determines the way the value for the element is retrieved 
 * from the bound data.
 */
ppc.spinner = function(struct, tagName){
    this.$init(tagName || "spinner", ppc.NODE_VISIBLE, struct);
    
    this.max     = 64000;
    this.min     = -64000;
    this.focused = false;
    this.value   = 0;
    
    this.realtime = false;
};

(function() {
    this.implement(
        //#ifdef __WITH_DATAACTION
        ppc.DataAction
        //#endif
        //#ifdef __WITH_XFORMS
        //,ppc.XForms
        //#endif
    );

    this.$supportedProperties.push("width", "value", "max", "min", "caption", "realtime");

    this.$booleanProperties["realtime"] = true;

    this.$propHandlers["value"] = function(value) {
        value = parseInt(value) || 0;
        
        this.value = this.oInput.value = (value > this.max
            ? this.max
            : (value < this.min
                ? this.min
                : value));
    };

    this.$propHandlers["min"] = function(value) {
        if (!(value = parseInt(value))) return;
        this.min = value;
        if (value > this.value)
            this.change(value);
    };

    this.$propHandlers["max"] = function(value) {
        if (!(value = parseInt(value))) return;
        this.max = value;

        if (value < this.value)
            this.change(value);
    };

    /* ********************************************************************
     PUBLIC METHODS
     *********************************************************************/

    //#ifdef __WITH_CONVENIENCE_API

    /**
     * Sets the value of this element. This should be one of the values
     * specified in the `values` attribute.
     * @param {String} value The new value of this element
     */
    this.setValue = function(value) {
       this.setProperty("value", value, false, true);
    };

    /**
     * Returns the current value of this element.
     * @return {String} The current element value
     */
    this.getValue = function() {
        return this.value;
    };

    /**
     * Increments the spinner by one.
     */
    this.increment = function() {
        this.change(parseInt(this.oInput.value) + 1);
    };

    /**
     * Decrements the spinner by one.
     */    
    this.decrement = function() {
        this.change(parseInt(this.oInput.value) - 1);
    };
    
    //#endif

    this.$enable = function() {
        this.oInput.disabled = false;
        this.$setStyleClass(this.oInput, "", ["inputDisabled"]);
    };

    this.$disable = function() {
        this.oInput.disabled = true;
        this.$setStyleClass(this.oInput, "inputDisabled");
    };

    this.$focus = function(e) {
        if (!this.$ext || this.focused) //this.disabled || 
            return;

        //#ifdef __WITH_WINDOW_FOCUS
        if (ppc.hasFocusBug)
            ppc.sanitizeTextbox(this.oInput);
        //#endif

        this.focused = true;
        this.$setStyleClass(this.oInput, "focus");
        this.$setStyleClass(this.$buttonPlus, "plusFocus");
        this.$setStyleClass(this.$buttonMinus, "minusFocus");
        
        if (this.oLeft)
            this.$setStyleClass(this.oLeft, "leftFocus");
    };

    this.$blur = function(e) {
        if (!this.$ext && !this.focused)
            return;

        this.$setStyleClass(this.oInput, "", ["focus"]);
        this.$setStyleClass(this.$buttonPlus, "", ["plusFocus"]);
        this.$setStyleClass(this.$buttonMinus, "", ["minusFocus"]);
        
        if (this.oLeft)
            this.$setStyleClass(this.oLeft, "" ["leftFocus"]);
        
        this.setValue(this.oInput.value);
        
        this.focused = false;
    };

    /* ***********************
     Keyboard Support
     ************************/
    //#ifdef __WITH_KEYBOARD
    this.addEventListener("keydown", function(e) {
        var key = e.keyCode,

        keyAccess = (key < 8 || (key > 9 && key < 37 && key !== 12)
            || (key > 40 && key < 46) || (key > 46 && key < 48)
            || (key > 57 && key < 96) || (key > 105 && key < 109 && key !== 107)
            || (key > 109 && key !== 189));

        if (keyAccess)
            return false;
           
        switch(key) {
            case 38://Arrow up
                this.increment();
                break;
            case 40://Arrow down
                this.decrement();
                break;
        }
    }, true);

    this.addEventListener("keyup", function(e) {
        //this.setValue(this.oInput.value);
    }, true);
    //#endif
    
    this.increment = function() {
        this.change(parseInt(this.oInput.value) + 1);
    };
    
    this.decrement = function() {
        this.change(parseInt(this.oInput.value) - 1);
    };
    
    /**
     * @event click     Fires when the user presses a mousebutton while over this element and then lets the mousebutton go. 
     */
    /**
     * @event mouseup   Fires when the user lets go of a mousebutton while over this element. 
     */
    /**
     * @event mousedown Fires when the user presses a mousebutton while over this element. 
     */
    this.$draw = function() {
        var _self = this;

        //Build Main Skin
        this.$ext = this.$getExternal(null, null, function(oExt) {
            oExt.setAttribute("onmousedown",
                'if (!this.host.disabled) \
                    this.host.dispatchEvent("mousedown", {htmlEvent : event});');
            oExt.setAttribute("onmouseup",
                'if (!this.host.disabled) \
                    this.host.dispatchEvent("mouseup", {htmlEvent : event});');
            oExt.setAttribute("onclick",
                'if (!this.host.disabled) \
                    this.host.dispatchEvent("click", {htmlEvent : event});');
        });

        this.$int         = this.$getLayoutNode("main", "container",   this.$ext);
        this.oInput       = this.$getLayoutNode("main", "input",       this.$ext);
        this.$buttonPlus  = this.$getLayoutNode("main", "buttonplus",  this.$ext);
        this.$buttonMinus = this.$getLayoutNode("main", "buttonminus", this.$ext);
        this.oLeft = this.$getLayoutNode("main", "left", this.$ext);

        //#ifdef __WITH_WINDOW_FOCUS
        ppc.sanitizeTextbox(this.oInput);
        //#endif

        var timer,
            doc = (!document.compatMode || document.compatMode == 'CSS1Compat')
                ? document.html : document.body,
            z   = 0;

        /* Setting start value */
        this.oInput.value = this.value;

        this.oInput.onmousedown = function(e) {
            if (_self.disabled)
                return;
            
            e = e || window.event;

            clearTimeout(timer);

            var newval,
                value = parseInt(this.value) || 0,
                step  = 0,
                cy    = e.clientY,
                ot    = _self.$int.offsetTop, ol = _self.$int.offsetLeft,
                ow    = _self.$int.offsetWidth, oh = _self.$int.offsetHeight,
                func  = function() {
                    clearTimeout(timer);
                    timer = $setTimeout(func, 10);
                    if (!step)
                        return;

                    newval = value + step;
                    if (newval <= _self.max && newval >= _self.min) {
                        value += step;
                        value = Math.round(value);
                        _self.oInput.value = value;
                        
                        if (_self.realtime)
                            _self.change(value);
                    }
                    else {
                        _self.oInput.value = step < 0
                            ? _self.min
                            : _self.max;
                    }
                };
            func();

            function calcStep(e) {
                e = e || window.event;
                var x = e.pageX || e.clientX + (doc ? doc.scrollLeft : 0),
                    y = e.pageY || e.clientY + (doc ? doc.scrollTop  : 0),
                    nrOfPixels = cy - y;

                if ((y > ot && x > ol) && (y < ot + oh && x < ol + ow)) {
                    step = 0;
                    return;
                }

                step = Math.pow(Math.min(200, Math.abs(nrOfPixels)) / 10, 2) / 10;
                if (nrOfPixels < 0)
                    step = -1 * step;
            }
            
            document.onmousemove = calcStep;

            document.onmouseup = function(e) {
                clearTimeout(timer);

                var value = parseInt(_self.oInput.value);

                if (value != _self.value)
                    _self.change(value);
                document.onmousemove = document.onmouseup = null;
            };
        };

        /* Fix for mousedown for IE */
        var buttonDown = false;
        this.$buttonPlus.onmousedown = function(e) {
            if (_self.disabled)
                return;
            
            e = e || window.event;
            buttonDown = true;

            var value = (parseInt(_self.oInput.value) || 0) + 1,
                func  = function() {
                    clearTimeout(timer);
                    timer = $setTimeout(func, 50);
                    z++;
                    value += Math.pow(Math.min(200, z) / 10, 2) / 10;
                    value = Math.round(value);

                    _self.oInput.value = value <= _self.max
                        ? value
                        : _self.max;
                    
                    if (_self.realtime)
                       _self.change(value <= _self.max ? value : _self.max);
                };

            ppc.setStyleClass(this, "plusDown", ["plusHover"]);

            func();
        };

        this.$buttonMinus.onmousedown = function(e) {
            if (_self.disabled)
                return;
            
            e = e || window.event;
            buttonDown = true;

            var value = (parseInt(_self.oInput.value) || 0) - 1,
                func  = function() {
                    clearTimeout(timer);
                    timer = $setTimeout(func, 50);
                    z++;
                    value -= Math.pow(Math.min(200, z) / 10, 2) / 10;
                    value = Math.round(value);

                    _self.oInput.value = value >= _self.min
                        ? value
                        : _self.min;
                    
                    if (_self.realtime)
                       _self.change(value >= _self.min ? value : _self.min);
                };

            ppc.setStyleClass(this, "minusDown", ["minusHover"]);

            func();
        };

        this.$buttonMinus.onmouseout = function(e) {
            if (_self.disabled)
                return;
            
            clearTimeout(timer);
            z = 0;

            var value = parseInt(_self.oInput.value);

            if (value != _self.value)
                _self.change(value);

            ppc.setStyleClass(this, "", ["minusHover"]);

            if (!_self.focused)
               _self.$blur(e);
        };

        this.$buttonPlus.onmouseout  = function(e) {
            if (_self.disabled)
                return;
            
            clearTimeout(timer);
            z = 0;

            var value = parseInt(_self.oInput.value);

            if (value != _self.value)
                _self.change(value);

            ppc.setStyleClass(this, "", ["plusHover"]);

            if (!_self.focused)
               _self.$blur(e);
        };

        this.$buttonMinus.onmouseover = function(e) {
            if (_self.disabled)
                return;
                
            ppc.setStyleClass(this, "minusHover");
        };

        this.$buttonPlus.onmouseover  = function(e) {
            if (_self.disabled)
                return;
                
            ppc.setStyleClass(this, "plusHover");
        };

        this.$buttonPlus.onmouseup = function(e) {
            if (_self.disabled)
                return;
            
            e = e || event;
            //e.cancelBubble = true;
            ppc.cancelBubble(e, this);

            ppc.setStyleClass(this, "plusHover", ["plusDown"]);

            clearTimeout(timer);
            z = 0;

            var value = parseInt(_self.oInput.value);

            if (!buttonDown) {
                value++;
                _self.oInput.value = value;
            }
            else {
                buttonDown = false;
            }

            if (value != _self.value)
                _self.change(value);
        };

        this.$buttonMinus.onmouseup = function(e) {
            if (_self.disabled)
                return;
            
            e = e || event;
            //e.cancelBubble = true;
            ppc.cancelBubble(e, this);

            ppc.setStyleClass(this, "minusHover", ["minusDown"]);

            clearTimeout(timer);
            z = 0;

            var value = parseInt(_self.oInput.value);

            if (!buttonDown) {
                value--;
                _self.oInput.value = value;
            }
            else {
                buttonDown = false;
            }


            if (value != _self.value)
                _self.change(value);
        };

        this.oInput.onselectstart = function(e) {
            e = e || event;
            e.cancelBubble = true;
        };

        this.oInput.host = this;
    };

    this.$destroy = function() {
        this.oInput.onkeypress =
        this.oInput.onmousedown =
        this.oInput.onkeydown =
        this.oInput.onkeyup =
        this.oInput.onselectstart =
        this.$buttonPlus.onmouseover =
        this.$buttonPlus.onmouseout =
        this.$buttonPlus.onmousedown =
        this.$buttonPlus.onmouseup =
        this.$buttonMinus.onmouseover =
        this.$buttonMinus.onmouseout =
        this.$buttonMinus.onmousedown =
        this.$buttonMinus.onmouseup = null;
    };
    
    // #ifdef __ENABLE_UIRECORDER_HOOK
    this.$getActiveElements = function() {
        // init $activeElements
        if (!this.$activeElements) {
            this.$activeElements = {
                oInput          : this.oInput,
                $buttonPlus     : this.$buttonPlus,
                $buttonMinus    : this.$buttonMinus
            }
        }

        return this.$activeElements;
    }
    //#endif

// #ifdef __WITH_DATABINDING
}).call(ppc.spinner.prototype = new ppc.StandardBinding());
/* #else
}).call(ppc.spinner.prototype = new ppc.Presentation());
#endif */

ppc.aml.setElement("spinner", ppc.spinner);
// #endif
